package operation

import (
	"context"
	"sync"
	"time"

	"git.perx.ru/perxis/perxis-go/pkg/auth"
	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"google.golang.org/protobuf/proto"
)

const (
	DefaultRetention       = time.Hour
	DefaultCleanupInterval = time.Minute
)

// Service интерфейс сервиса для работы с продолжительными операциями
type Service interface {

	// Create создает новую операцию и запускает ее в фоне
	Create(ctx context.Context, desc string, fn func(ctx context.Context) (proto.Message, error)) (*Operation, error)

	// Set сохраняет операцию в хранилище
	Set(ctx context.Context, op *Operation) error

	// Get возвращает операцию по ее идентификатору
	Get(ctx context.Context, id string) (*Operation, error)

	// Cancel отменяет операцию по ее идентификатору
	Cancel(ctx context.Context, id string) (*Operation, error)

	// Cleanup удаляет просроченные операции
	Cleanup()
}

type service struct {
	ops             map[string]*Operation // Registry of operations
	mu              sync.RWMutex          // Mutex for operations registry
	lastCleanup     time.Time             // Last time when cleanup was performed
	cleanupInterval time.Duration         // Cleanup interval
}

func NewService(cleanupInterval time.Duration) Service {
	return &service{
		ops:             make(map[string]*Operation),
		mu:              sync.RWMutex{},
		cleanupInterval: cleanupInterval,
	}
}

func NewDefaultService() Service {
	return &service{
		ops:             make(map[string]*Operation),
		mu:              sync.RWMutex{},
		cleanupInterval: DefaultCleanupInterval,
	}
}

func (s *service) Cleanup() {
	s.mu.Lock()
	defer s.mu.Unlock()
	for id, op := range s.ops {
		if op.IsExpired() {
			delete(s.ops, id)
		}
	}
}

func (s *service) CleanupIfNeeded() {
	if time.Since(s.lastCleanup) > s.cleanupInterval {
		s.lastCleanup = time.Now()
		s.Cleanup()
	}
}

// Set сохраняет операцию в хранилище
func (s *service) Set(_ context.Context, op *Operation) error {
	if op == nil || op.Id() == "" {
		return errors.New("invalid operation")
	}

	if op.retention == 0 {
		op.retention = DefaultRetention
	}

	s.mu.Lock()
	s.ops[op.Id()] = op
	s.mu.Unlock()

	s.CleanupIfNeeded()
	return nil
}

// Get возвращает операцию по ее идентификатору
func (s *service) Get(_ context.Context, id string) (*Operation, error) {
	s.mu.RLock()
	defer s.mu.RUnlock()
	op, ok := s.ops[id]
	if !ok {
		return nil, errors.New("operation not found")
	}
	return op, nil
}

// Cancel отменяет операцию по ее идентификатору
func (s *service) Cancel(_ context.Context, id string) (*Operation, error) {
	s.mu.RLock()
	defer s.mu.Unlock()
	op, ok := s.ops[id]
	if !ok {
		return nil, errors.New("operation not found")
	}

	if op.cancelFunc != nil {
		op.cancelFunc()
	}

	return op, nil
}

func (s *service) Create(ctx context.Context, desc string, fn func(ctx context.Context) (proto.Message, error)) (*Operation, error) {
	principal := auth.GetPrincipal(ctx)
	op := New(&Proto{
		Description: desc,
		CreatedBy:   principal.GetID(ctx),
	})

	// Создаем новый контекст для фоновой операции без Deadline
	opCtx, cancel := context.WithCancel(auth.WithPrincipal(context.Background(), principal))
	op.SetCancelFunc(cancel)
	_ = s.Set(opCtx, op)

	go func() {
		response, err := fn(opCtx)
		if err != nil {
			op.SetError(err)
		}
		if response != nil {
			_ = op.SetResponse(response)
		}
		op.SetModifiedAt(time.Now())
		op.SetDone(true)
		_ = s.Set(opCtx, op)
	}()

	return op, nil
}
